#ifndef _ARV_B_LIBDADOS_H_
#define _ARV_B_LIBDADOS_H_

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#ifndef bool 
typedef enum _bool {
  false, true
} bool;
#endif

#define NMAXREGS 511 /* Este  o nmero maximo (mxima ordem possvel)
			de registros de um n. */

typedef struct arvb {  /* Registro que representa n em rvore B. Sua
			  rvore deve ser um ndice primrio e,
			  portanto, no poder ter chaves
			  repetidas. */
  int ordem;  /* nmero mximo de registros em um n da arvore que
		 est sendo criada */
  int nregs;  /* nmero de registros no n */
  int chave[NMAXREGS+1]; /* vetor contendo as chaves dos registros em
			    ordem crescente. Uma posio extra 
			    necessria para acomodar o overflow
			    (nmero de chaves maior do que o
			    possvel). */
  struct arvb *filha[NMAXREGS+2]; /* Vetor contendo apontadores para
			          as rvores filhas do n. Uma posio
			          extra  necessria para acomodar o
			          overflow (nmero de filhas maior do
			          que o possvel). */
} ArvB, NoArvB;

NoArvB *CriaNoArvB(int ordem);
/* Cria um n da arvore B com a ordem passada como parametro. Devolve
   NULL caso a ordem seja maior do que o tamanho maximo especificado,
   ou menor que 2.
*/

bool DestroiArvB(ArvB **arvore);
/* Destroi a Arvore B. O apontador da arvore deve retornar NULL.
   A funo retorna true se tudo deu certo, ou false no caso contrrio.
*/

bool BuscaChaveArvB(ArvB *arvore, int chave, NoArvB **no, int *pos);
/* Busca a chave de um registro na arvore B.  A funo deve retornar
   false se o valor no existir na rvore e abortar se a rvore no
   existir. Caso a chave esteja na rvore, a funo dever retornar o
   apontador para o n que contm a chave e a sua posio no vetor de
   chaves em "pos".
*/

bool InsereArvB(ArvB **arvore, int chave);
/* Insere a chave na rvore B. Retorna true se deu tudo certo, retorna
   false se no foi possvel a insero (e.g., se o valor j existe).
*/

bool RemoveArvB(ArvB **arvore, int chave);
/* Remove a chave da arvore B, retornando false se a chave no for
   encontrada para remoo, abortando caso a rvore no exista, e
   retornando true caso a remoo tenha sido bem sucedida. */

void ImprimeArvBPreOrdem(ArvB *arv, int nivel);
void ImprimeArvB(ArvB *ab, int nivel, int filha);

#endif





